//---------------------------------------------------------------------------
// Includes
#include "functionfs.h"
#include "main.h"
#include "usbgadget.h"
#include <linux/usb/ch9.h>			// Kapitel 9 des USB-Standards
#include <linux/usb/functionfs.h>	// FunctionFS
#include <errno.h>					// Fehlercodes
#include <sys/stat.h>				// mkdir
#include <sys/mount.h>				// Filesysteme mounten
#include <stddef.h>					// NULL
#include <fcntl.h>					// open
#include <poll.h>					// poll
#include <pthread.h>				// PThreads
#include <stdio.h>
#include <string.h>					// strerror
#include <unistd.h>					// write
#include <malloc.h>
#include <signal.h>					// pthread_kill

#include "defines.h"

//---------------------------------------------------------------------------
// Private Makros
#define MOUNTPATH_STR		"/ffs"		// Mount-Pfad für das FunctionFS

#define ENDPOINT_CNT	2		// Anzahl der Endpoints des Gerätes (inkl. ep0)

//---------------------------------------------------------------------------
// Private Typen
typedef void* (*EP_Thread)(void* arg);

//---------------------------------------------------------------------------
// Private Funktionen
static int _EP0_Init(void);
static void *_EP0_Thread(void* arg);

static int _EPx_Init(void);
static int _EPx_DeInit(void);
static int _EPx_Read(struct pollfd ep_poll[2], size_t poll_cnt, char* buffer, size_t buffer_size);
static void *_EP1_Thread(void* arg);
static int _Setup_Handler(void);

//---------------------------------------------------------------------------
// Private Variablen

static const struct {
	struct usb_functionfs_descs_head_v2 header;
	__le32 fs_count;
	__le32 hs_count;
	__le32 os_count;
	struct {
		struct usb_interface_descriptor intf;
		struct usb_endpoint_descriptor_no_audio ep1;
	} __attribute__((packed)) fs_descs, hs_descs;
	struct {
		struct usb_os_desc_header header;
		struct usb_ext_compat_desc function;
	} __attribute__((packed)) os_desc;
} __attribute__((packed)) ep0_desc =
		{
			.header = {
				.magic  = __cpu_to_le32(FUNCTIONFS_DESCRIPTORS_MAGIC_V2),
				.length = __cpu_to_le32(sizeof ep0_desc),
				.flags  = __cpu_to_le32(FUNCTIONFS_HAS_FS_DESC |
										FUNCTIONFS_HAS_HS_DESC |
										FUNCTIONFS_HAS_MS_OS_DESC),
			},
			.fs_count = __cpu_to_le32(ENDPOINT_CNT),
			.fs_descs = {
				.intf = {
					.bLength = sizeof(ep0_desc.fs_descs.intf),
					.bDescriptorType = USB_DT_INTERFACE,
					.bInterfaceNumber = 0,
					.bAlternateSetting = 0,
					.bNumEndpoints = ENDPOINT_CNT - 1,
					.bInterfaceClass = USB_CLASS_VENDOR_SPEC,
					.bInterfaceSubClass = 0,
					.bInterfaceProtocol = 0,
					.iInterface = 0,
				},
				.ep1 = {
					.bLength = sizeof(ep0_desc.fs_descs.ep1),
					.bDescriptorType = USB_DT_ENDPOINT,
					.bEndpointAddress = 1 | USB_DIR_OUT,
					.bmAttributes = USB_ENDPOINT_XFER_BULK,
					.wMaxPacketSize = 0,
					.bInterval = 0,
				},
			},
			.hs_count = __cpu_to_le32(ENDPOINT_CNT),
			.hs_descs = {
				.intf = {
					.bLength = sizeof(ep0_desc.fs_descs.intf),
					.bDescriptorType = USB_DT_INTERFACE,
					.bInterfaceNumber = 0,
					.bAlternateSetting = 0,
					.bNumEndpoints = ENDPOINT_CNT - 1,
					.bInterfaceClass = USB_CLASS_VENDOR_SPEC,
					.bInterfaceSubClass = 0,
					.bInterfaceProtocol = 0,
					.iInterface = 0,
				},
				.ep1 = {
					.bLength = sizeof(ep0_desc.hs_descs.ep1),
					.bDescriptorType = USB_DT_ENDPOINT,
					.bEndpointAddress = 1 | USB_DIR_OUT,
					.bmAttributes = USB_ENDPOINT_XFER_BULK,
					.wMaxPacketSize = __constant_cpu_to_le16(512),
					.bInterval = 0,
				},
			},
			.os_count = __cpu_to_le32(1),
			.os_desc = {
				.header = {
					.interface = 0,
					.dwLength = __cpu_to_le32(sizeof ep0_desc.os_desc),
					.bcdVersion = __cpu_to_le16(1),
					.wIndex = __cpu_to_le16(4),
					.bCount = 1,
					.Reserved = 0,
				},
				.function = {
					.bFirstInterfaceNumber = 0,
					.Reserved1 = 0,
					.CompatibleID = "WINUSB",
					.SubCompatibleID = {0},
					.Reserved2 = {0},
				},
			},
		};

//#define STR_INTERFACE_ "Source/Sink"
static const struct {
	struct usb_functionfs_strings_head header;
} __attribute__((packed)) strings = {
	.header = {
		.magic = __cpu_to_le32(FUNCTIONFS_STRINGS_MAGIC),
		.length = __cpu_to_le32(sizeof strings),
		.str_count = __cpu_to_le32(0),
		.lang_count = __cpu_to_le32(0),
	},
};

static struct {
	int fd;			// Datei des Endpoints
	int pipe[2];	// Pipe für Interthread-Kommunikation ([0] - Lesen, [1] - Schreiben)
	pthread_t id;	// Thread-ID
	EP_Thread func;	// Funktionspointer des EP-Handlers
} ep[2] = {	// Endpoint
		{-1, {-1, -1}, 0, _EP0_Thread},
		{-1, {-1, -1}, 0, _EP1_Thread},
};

//---------------------------------------------------------------------------
// Globale Variablen

//---------------------------------------------------------------------------

int FunctionFS_Init(void)
{
	int ret = -EINVAL;

	// Mount-Ordner erstellen
	ret = mkdir(MOUNTPATH_STR, S_IRWXU | S_IRWXG | S_IRWXO);
	if ((ret == 0) || (errno == EEXIST))
	{
		// Dateisystem mounten
		ret = mount(USBGADGET_FUNCTION_INST_NAME_STR, MOUNTPATH_STR, "functionfs", 0, NULL);
		if (ret == 0)
		{
			if (bverbose == true)
				fprintf(stdout, "Filesystem successfully mounted\n");

			// EP0 initialisieren
			ret = _EP0_Init();
			if (ret >= 0)
				// EP0-Handler starten
				pthread_create(&ep[0].id, NULL, ep[0].func, NULL);

		}
		else
		{
			fprintf(stderr, "Filesystem couldn't be mounted\n");
			fprintf(stderr, "Error: %s\n", strerror(errno));
		}
	}
	else
	{
		fprintf(stderr, "Directory couldn't be created\n");
		fprintf(stderr, "Error: %s\n", strerror(errno));
	}

	return ret;
}

void FunctionFS_Main(void)
{
	// auf das Ende von ep0 warten
	pthread_join(ep[0].id, NULL);
}

static int _EP0_Init(void)
{
	int ret = -EINVAL;

	// EP0 wird initialisiert:
	// EP0 öffnen und Deskriptoren & Strings schreiben
	ep[0].fd = open(MOUNTPATH_STR"/ep0", O_RDWR);
	if (ep[0].fd < 0)
	{
		perror(MOUNTPATH_STR"/ep0");
	}
	else if (write(ep[0].fd, &ep0_desc, sizeof(ep0_desc)) < 0)
	{
		perror("write dev descriptors");
	}
	else if (write(ep[0].fd, &strings, sizeof strings) < 0)
	{
		perror("write dev strings");
	}
	else
	{
		ret = 0;

		if (bverbose == true)
		{
			fprintf(stdout, "EP0 initialised\n");
		}
	}

	return ret;
}

static void *_EP0_Thread(void* arg)
{
	static const char *const names[] = {
		[FUNCTIONFS_BIND] = "BIND",
		[FUNCTIONFS_UNBIND] = "UNBIND",
		[FUNCTIONFS_ENABLE] = "ENABLE",
		[FUNCTIONFS_DISABLE] = "DISABLE",
		[FUNCTIONFS_SETUP] = "SETUP",
		[FUNCTIONFS_SUSPEND] = "SUSPEND",
		[FUNCTIONFS_RESUME] = "RESUME",
	};

	struct pollfd		ep0_poll;
	ep0_poll.fd = ep[0].fd;
	ep0_poll.events = POLLIN;

	for(;;)
	{
		struct usb_functionfs_event events[5];		// todo: warum 5?
		bool bEndThread = false;

		int tmp = poll(&ep0_poll, 1, -1);
		if (tmp < 0) {
			perror("poll");
			break;
		}

		int foo = ioctl(ep0_poll.fd, FUNCTIONFS_FIFO_STATUS);
		if (foo < 0)	perror("ep0 ioctl");
		else			fprintf(stdout, "%i\n", foo);

		int ret = read(ep[0].fd, &events, sizeof(events));
		if (ret < 0)
		{
			// Lesefehler -> Thread beenden
			perror("ep0 read");
			bEndThread = true;
		}
		else
		{
			// Anzahl der ausgelesenen Events
			int event_cnt = ret / sizeof(events[0]);
			if (bverbose == true)
			{
				printf("%d event(s)\n", event_cnt);
			}

			// die Events auswerten
			for (int i = 0; i < event_cnt; i++)
			{
				unsigned char type = events[i].type;
				if (bverbose == true)
				{
					if (type < elemof(names))	printf("%s\n", names[type]);
					else						printf("unknown event %d\n", type);
				}

				switch (type)
				{
					case FUNCTIONFS_ENABLE:
					{
						_EPx_Init();
						break;
					}
					case FUNCTIONFS_DISABLE:
					{
						_EPx_DeInit();
						break;
					}
					case FUNCTIONFS_SETUP:
					{
						_Setup_Handler();
						break;
					}
					case FUNCTIONFS_BIND:
					case FUNCTIONFS_UNBIND:
					case FUNCTIONFS_SUSPEND:
					case FUNCTIONFS_RESUME:
					default:
					{
						break;
					}
				}
			}
		}

		if (bEndThread == true)
		{
			break;
		}
	}

	if (ep[0].fd >= 0)
	{
		close(ep[0].fd);
	}

	if (bverbose == true)
	{
		fprintf(stdout, "EP0-Thread beendet\n");
	}

	// Thread beenden
	pthread_exit(NULL);
}

static int _EPx_Init(void)
{
	int ret = -EINVAL;

	if (bverbose == true)
	{
		fprintf(stdout, "Start Endpoints\n");
	}

	for (unsigned i = 1; i < elemof(ep); i++)
	{
		char* path = (char*)malloc(strlen(MOUNTPATH_STR) + strlen("/epx") + 1);
		if (path != NULL)
		{
			// Dateipfad des Endpoints bauen
			sprintf(path, "%s/ep%u", MOUNTPATH_STR, i);

			// Endpoint öffnen
			if ((ep[i].fd = open(path, O_RDWR)) < 0)
			{
				perror(path);
			}
			// Pipe einrichten
			else if (pipe(ep[i].pipe) < 0)
			{
				perror("Pipe");
			}
			// Thread für Endpoint starten
			else if ((pthread_create(&ep[i].id, NULL, ep[i].func, NULL)) < 0)
			{
				fprintf(stderr, "Cannot create thread for ep%u\n", i);
			}

			free(path);
		}
	}

	return ret;
}

static int _EPx_DeInit(void)
{
	int ret = -EINVAL;

	if (bverbose == true)
	{
		fprintf(stdout, "Stop Endpoints\n");
	}

	for (unsigned i = 1; i < elemof(ep); i++)
	{
		// prüfen, ob der Thread noch läuft
		if (pthread_kill(ep[i].id, 0) == 0)
		{
			// Thread von EPx beenden
			char c = 0x01;
			write(ep[i].pipe[1], &c, sizeof(c));

			// warten, bis sich dieser Thread beendet
			pthread_join(ep[i].id, NULL);
		}

		// Endpoint & Pipe schließen
		close(ep[i].fd);
		ep[i].fd = -1;
		close(ep[i].pipe[0]);
		ep[i].pipe[0] = -1;
		close(ep[i].pipe[1]);
		ep[i].pipe[1] = -1;
	}

	return ret;
}

static int _EPx_Read(struct pollfd ep_poll[2], size_t poll_cnt, char* buffer, size_t buffer_size)
{
	int ret = 0;

	// auf I/O-Vorgang vom EP oder der Pipe warten (blockiert den Thread)
	int tmp = poll(ep_poll, poll_cnt, -1);
	if (tmp < 0) {
		perror("poll");
		ret = -1;
	}
	else if (tmp > 0)
	{
		// Pipe prüfen -> Thread beenden
		if (ep_poll[0].revents != 0)
		{
			if (bverbose == true)
			{
				fprintf(stdout, "Close pipe\n");
			}
			ret = -1;
		}
		// EP prüfen
		else if (ep_poll[1].revents != 0)
		{
			int foo = ioctl(ep_poll[1].fd, FUNCTIONFS_FIFO_STATUS);
			if (foo < 0)	perror("epx ioctl");
			else			fprintf(stdout, "%i\n", foo);

			ep_poll[1].revents = 0;
			if (read(ep_poll[1].fd, buffer, buffer_size) < 0)
			{
				// Lesefehler -> Thread beenden
				perror("ep read");
				//ret = -1;
			}
		}
	}

	return ret;
}

static void *_EP1_Thread(void* arg)
{
	if (bverbose == true)
	{
		fprintf(stdout, "EP1-Thread gestartet\n");
	}

	struct pollfd ep_poll[2] = {
			{ep[1].pipe[0], POLLIN, POLLIN},
			{ep[1].fd, POLLIN, POLLIN},
	};

	for(;;)
	{
		char buffer[8 * 1024];
		bool bEndThread = false;

		int ret = _EPx_Read(ep_poll, elemof(ep_poll), buffer, sizeof(buffer));
		if (ret < 0)
		{
			bEndThread = true;
		}
		else
		{
			// todo: Daten verarbeiten
		}

		if (bEndThread == true)
		{
			break;
		}
	}

	if (bverbose == true)
	{
		fprintf(stdout, "EP1-Thread beendet\n");
	}

	// Thread beenden
	pthread_exit(NULL);
}

static int _Setup_Handler(void)
{
	int ret = -EINVAL;

	// todo Befehle verarbeiten

	return ret;
}
