//---------------------------------------------------------------------------
// Includes
#include "usbgadget.h"
#include "main.h"
#include <errno.h>					// Fehlercodes
#include <usbgx/usbg/usbg.h>		// ConfigFS-Lib
#include <linux/usb/ch9.h>			// Kapitel 9 des USB-Standards

//---------------------------------------------------------------------------
// Private Makros
#define GADGETFS_MOUNTPATH_STR		"/sys/kernel/config"	// Mount-Pfad für das ConfigFS
#define GADGET_NAME_STR				"g1"					// Name des Gadgets

#define VENDORID		0x0D1D	// USB Vendor ID
#define PRODUCTID		0x0020	// USB Produkt ID

#define MANU_STR	"Hersteller"	// Hersteller
#define PRODUCT_STR	"Produkt"		// Gerätebezeichnung
#define SERIAL_STR	"0123456789"	// Seriennummer

//---------------------------------------------------------------------------
// Private Typen

//---------------------------------------------------------------------------
// Private Funktionen
static void _USBgadget_PrintError(usbg_error usbg_ret);

//---------------------------------------------------------------------------
// Private Variablen
// Geräteattribute (Device Descriptor)
static struct usbg_gadget_attrs g_attr =
{
	.bcdUSB = 0x0200,			// Unterstützte USB-Version (2.0)
	.bDeviceClass =	USB_CLASS_PER_INTERFACE,	// wird vom Interface Descriptor beschrieben
	.bDeviceSubClass = 0x00,
	.bDeviceProtocol = 0x00,
	.bMaxPacketSize0 = 64, 		// Maximale Paketgröße für ep0
	.idVendor = VENDORID,
	.idProduct = PRODUCTID,
	.bcdDevice = 0x0100, /* Verson of device */
};

// Geräte-Zeichenketten (Device Descriptor)
static struct usbg_gadget_strs g_str =
{
	.manufacturer	= (char*)MANU_STR,
	.product		= (char*)PRODUCT_STR,
	.serial			= (char*)SERIAL_STR
};

// Konfiguration-Zeichenkette (Configuration Descriptor)
static struct usbg_config_strs c_str =
{
	.configuration = (char*)"Configuration 1"
};

struct usbg_gadget_os_descs g_os_desc = {
	.use = true,
	.b_vendor_code = 0x01,
	.qw_sign = "MSFT100",
};

// Konfiguration-Attribute (Configuration Descriptor)
static struct usbg_config_attrs c_attr =
{
	.bmAttributes = 0x80,
	.bMaxPower = 0xFA,
};

static usbg_state *s;	// Status des Gadgets im System
static usbg_gadget *g;	// das Gadget

//---------------------------------------------------------------------------
// Globale Variablen

//---------------------------------------------------------------------------

int USBgadget_Init(void)
{
	int ret = -EINVAL;

	usbg_function *f_ffs;	// Funktion des Gadgets
	usbg_config *c;			// Konfiguration des Gadgets

	// Lib initialisieren und GadgetFS mounten
	int usbg_ret = usbg_init(GADGETFS_MOUNTPATH_STR, &s);
	if (usbg_ret == USBG_SUCCESS)
	{
		// evtl. bereits vorhandenes Gadget entfernen
		USBgadget_Reset();

		// neues Gadget erstellen und Device Descriptor setzen
		if ((usbg_ret = usbg_create_gadget(s, GADGET_NAME_STR, &g_attr, &g_str, &g)) != USBG_SUCCESS)
		{
			fprintf(stderr, "Error on create gadget\n");
			_USBgadget_PrintError((usbg_error)usbg_ret);
		}
		// neue Funktion erstellen
		else if ((usbg_ret = usbg_create_function(g, USBG_F_FFS, USBGADGET_FUNCTION_INST_NAME_STR, NULL, &f_ffs)) != USBG_SUCCESS)
		{
			fprintf(stderr, "Error creating ffs function\n");
			_USBgadget_PrintError((usbg_error)usbg_ret);
		}
		// neue Konfiguration erstellen
		else if ((usbg_ret = usbg_create_config(g, 1, NULL, &c_attr, NULL, &c)) != USBG_SUCCESS)
		{
			fprintf(stderr, "Error creating config\n");
			_USBgadget_PrintError((usbg_error)usbg_ret);
		}
		// Funktion an Konfiguration binden
		else if ((usbg_ret = usbg_add_config_function(c, "ffs."USBGADGET_FUNCTION_INST_NAME_STR, f_ffs)) != USBG_SUCCESS)
		{
			fprintf(stderr, "Error adding ffs\n");
			_USBgadget_PrintError((usbg_error)usbg_ret);
		}
		// OS Descriptor einrichten
		else if ((usbg_ret = usbg_set_gadget_os_descs(g, &g_os_desc)) != USBG_SUCCESS)
		{
			fprintf(stderr, "Error creating os descriptor\n");
			_USBgadget_PrintError((usbg_error)usbg_ret);
		}
		// OS Descriptor an Konfiguration binden
		else if ((usbg_ret = usbg_set_os_desc_config(g, c)) != USBG_SUCCESS)
		{
			fprintf(stderr, "Error adding os descriptor\n");
			_USBgadget_PrintError((usbg_error)usbg_ret);
		}
		else
		{
			if (bverbose == true)
				fprintf(stdout, "Gadget has been created\n");
			ret = 0;
		}

		if (ret != 0)
		{
			usbg_cleanup(s);
		}
	}
	else
	{
		fprintf(stderr, "Error on USB gadget init\n");
		_USBgadget_PrintError((usbg_error)usbg_ret);
	}

	return ret;
}

int USBgadget_Start(void)
{
	int ret = -EINVAL;

	// Gadget wird mit dem erstbesten UDC (USB Device Controller) gestartet
	int usbg_ret = usbg_enable_gadget(g, DEFAULT_UDC);
	if (usbg_ret == USBG_SUCCESS)
	{
		ret = 0;
	}
	else
	{
		fprintf(stderr, "Error on USB gadget enable\n");
		_USBgadget_PrintError((usbg_error)usbg_ret);
	}

	return ret;
}

int USBgadget_Reset(void)
{
	int ret = -EINVAL;

	g = usbg_get_gadget(s, GADGET_NAME_STR);
	if (g != NULL)
	{
		usbg_disable_gadget(g);
		usbg_rm_gadget(g, USBG_RM_RECURSE);
	}

	return ret;
}

static void _USBgadget_PrintError(usbg_error usbg_ret)
{
	fprintf(stderr, "Error: %s : %s\n", usbg_error_name(usbg_ret), usbg_strerror(usbg_ret));
}
